@@ -26,10 +26,9 @@
 
 void wcmSetPressureCurve(WacomDevicePtr pDev, int x0, int y0,
 	int x1, int y1);
-int wcmFilterIntuos(WacomCommonPtr common, WacomChannelPtr pChannel,
-	WacomDeviceStatePtr ds);
 int wcmFilterCoord(WacomCommonPtr common, WacomChannelPtr pChannel,
 	WacomDeviceStatePtr ds);
+void wcmResetSampleCounter(const WacomChannelPtr pChannel);
 
 /****************************************************************************/
 #endif /* __XF86_WCMFILTER_H */
