@@ -1,5 +1,5 @@
 /*
- * Copyright 2009 Red Hat, Inc.
+ * Copyright 2009 - 2010 Red Hat, Inc.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -34,12 +34,13 @@
 #include <X11/Xlib.h>
 #include <X11/Xatom.h>
 #include <X11/extensions/XInput.h>
+#include <X11/extensions/Xrandr.h>
 #include <X11/XKBlib.h>
 
 #define TRACE(...) \
 	if (verbose) fprintf(stderr, "... " __VA_ARGS__)
 
-#define ArrayLength(a) (sizeof(a) / (sizeof((a)[0])))
+#define ArrayLength(a) ((unsigned int)(sizeof(a) / (sizeof((a)[0]))))
 
 static int verbose = False;
 
@@ -83,6 +84,7 @@ typedef struct _param
 	const char *prop_name;	/* property name */
 	const int prop_format;	/* property format */
 	const int prop_offset;	/* offset (index) into the property values */
+	const int prop_extra;   /* extra number of items after first one */
 	const unsigned int prop_flags;
 	void (*set_func)(Display *dpy, XDevice *dev, struct _param *param, int argc, char **argv); /* handler function, if appropriate */
 	void (*get_func)(Display *dpy, XDevice *dev, struct _param *param, int argc, char **argv); /* handler function for getting, if appropriate */
@@ -94,17 +96,17 @@ typedef struct _param
 
 /* get_func/set_func calls for special parameters */
 static void map_button(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
+static void map_wheels(Display *dpy, XDevice *dev, param_t* param, int argc, char **argv);
 static void set_mode(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
 static void get_mode(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
 static void get_presscurve(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
 static void get_button(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
 static void set_rotate(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
 static void get_rotate(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
-static void set_twinview(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
-static void get_twinview(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
 static void set_xydefault(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
 static void get_all(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
 static void get_param(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
+static void set_output(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
 
 static param_t parameters[] =
 {
@@ -375,16 +377,6 @@ static param_t parameters[] =
 		.prop_offset = 0,
 		.get_func = get_presscurve,
 	},
-	{
-		.name = "TwinView",
-		.desc = "Sets the mapping to TwinView horizontal/vertical/none. "
-		"Values = none, vertical, horizontal (default is none).",
-		.prop_name = WACOM_PROP_DISPLAY_OPTS,
-		.prop_format = 8,
-		.prop_offset = 1,
-		.get_func = get_twinview,
-		.set_func = set_twinview,
-	},
 	{
 		.name = "Mode",
 		.desc = "Switches cursor movement mode (default is absolute/on). ",
@@ -432,7 +424,7 @@ static param_t parameters[] =
 		"(default is 20). ",
 		.prop_name = WACOM_PROP_GESTURE_PARAMETERS,
 		.prop_format = 32,
-		.prop_offset = 0,
+		.prop_offset = 1,
 	},
 	{
 		.name = "TapTime",
@@ -440,7 +432,7 @@ static param_t parameters[] =
 		"(default is 250). ",
 		.prop_name = WACOM_PROP_GESTURE_PARAMETERS,
 		.prop_format = 32,
-		.prop_offset = 0,
+		.prop_offset = 2,
 	},
 	{
 		.name = "Capacity",
@@ -474,6 +466,7 @@ static param_t parameters[] =
 		.prop_name = WACOM_PROP_WHEELBUTTONS,
 		.prop_format = 8,
 		.prop_offset = 0,
+		.set_func = map_wheels,
 	},
 	{
 		.name = "RelWDn",
@@ -481,6 +474,7 @@ static param_t parameters[] =
 		.prop_name = WACOM_PROP_WHEELBUTTONS,
 		.prop_format = 8,
 		.prop_offset = 1,
+		.set_func = map_wheels,
 	},
 	{
 		.name = "AbsWUp",
@@ -488,6 +482,7 @@ static param_t parameters[] =
 		.prop_name = WACOM_PROP_WHEELBUTTONS,
 		.prop_format = 8,
 		.prop_offset = 2,
+		.set_func = map_wheels,
 	},
 	{
 		.name = "AbsWDn",
@@ -495,6 +490,7 @@ static param_t parameters[] =
 		.prop_name = WACOM_PROP_WHEELBUTTONS,
 		.prop_format = 8,
 		.prop_offset = 3,
+		.set_func = map_wheels,
 	},
 	{
 		.name = "StripLUp",
@@ -502,6 +498,7 @@ static param_t parameters[] =
 		.prop_name = WACOM_PROP_STRIPBUTTONS,
 		.prop_format = 8,
 		.prop_offset = 0,
+		.set_func = map_wheels,
 	},
 	{
 		.name = "StripLDn",
@@ -509,6 +506,7 @@ static param_t parameters[] =
 		.prop_name = WACOM_PROP_STRIPBUTTONS,
 		.prop_format = 8,
 		.prop_offset = 1,
+		.set_func = map_wheels,
 	},
 	{
 		.name = "StripRUp",
@@ -516,6 +514,7 @@ static param_t parameters[] =
 		.prop_name = WACOM_PROP_STRIPBUTTONS,
 		.prop_format = 8,
 		.prop_offset = 2,
+		.set_func = map_wheels,
 	},
 	{
 		.name = "StripRDn",
@@ -523,20 +522,7 @@ static param_t parameters[] =
 		.prop_name = WACOM_PROP_STRIPBUTTONS,
 		.prop_format = 8,
 		.prop_offset = 3,
-	},
-	{
-		.name = "TVResolution0",
-		.desc = "Sets MetaModes option for TwinView Screen 0. ",
-		.prop_name = WACOM_PROP_TWINVIEW_RES,
-		.prop_format = 32,
-		.prop_offset = 0,
-	},
-	{
-		.name = "TVResolution1",
-		.desc = "Sets MetaModes option for TwinView Screen 1. ",
-		.prop_name = WACOM_PROP_TWINVIEW_RES,
-		.prop_format = 32,
-		.prop_offset = 1,
+		.set_func = map_wheels,
 	},
 	{
 		.name = "RawFilter",
@@ -548,9 +534,9 @@ static param_t parameters[] =
 		.prop_flags = PROP_FLAG_BOOLEAN
 	},
 	{
-		.name = "ClickForce",
+		.name = "Threshold",
 		.desc = "Sets tip/eraser pressure threshold "
-		"(default is 409)",
+		"(default is 27)",
 		.prop_name = WACOM_PROP_PRESSURE_THRESHOLD,
 		.prop_format = 32,
 		.prop_offset = 0,
@@ -572,262 +558,6 @@ static param_t parameters[] =
 		.prop_format = 8,
 		.prop_offset = 2,
 	},
-	{
-		.name = "STopX0",
-		.desc = "Screen 0 left coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 0,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "STopY0",
-		.desc = "Screen 0 top coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 1,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "SBottomX0",
-		.desc = "Screen 0 right coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 2,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "SBottomY0",
-		.desc = "Screen 0 bottom coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 3,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "STopX1",
-		.desc = "Screen 1 left coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 4,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "STopY1",
-		.desc = "Screen 1 top coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 5,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "SBottomX1",
-		.desc = "Screen 1 right coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 6,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "SBottomY1",
-		.desc = "Screen 1 bottom coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 7,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "STopX2",
-		.desc = "Screen 2 left coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 8,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "STopY2",
-		.desc = "Screen 2 top coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 9,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "SBottomX2",
-		.desc = "Screen 2 right coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_offset = 10,
-		.prop_format = 32,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "SBottomY2",
-		.desc = "Screen 2 bottom coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 11,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "STopX3",
-		.desc = "Screen 3 left coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 12,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "STopY3",
-		.desc = "Screen 3 top coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 13,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "SBottomX3",
-		.desc = "Screen 3 right coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 14,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "SBottomY3",
-		.desc = "Screen 3 bottom coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 15,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "STopX4",
-		.desc = "Screen 4 left coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 16,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "STopY4",
-		.desc = "Screen 4 top coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 17,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "SBottomX4",
-		.desc = "Screen 4 right coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 18,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "SBottomY4",
-		.desc = "Screen 4 bottom coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 19,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "STopX5",
-		.desc = "Screen 5 left coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 20,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "STopY5",
-		.desc = "Screen 5 top coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 21,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "SBottomX5",
-		.desc = "Screen 5 right coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 22,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "SBottomY5",
-		.desc = "Screen 5 bottom coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 23,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "STopX6",
-		.desc = "Screen 6 left coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 24,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "STopY6",
-		.desc = "Screen 6 top coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 25,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "SBottomX6",
-		.desc = "Screen 6 right coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 26,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "SBottomY6",
-		.desc = "Screen 6 bottom coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 27,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "STopX7",
-		.desc = "Screen 7 left coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 28,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "STopY7",
-		"Screen 7 top coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 29,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "SBottomX7",
-		.desc = "Screen 7 right coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 30,
-		.prop_flags = PROP_FLAG_READONLY
-	},
-	{
-		.name = "SBottomY7",
-		.desc = "Screen 7 bottom coordinate in pixels. ",
-		.prop_name = WACOM_PROP_SCREENAREA,
-		.prop_format = 32,
-		.prop_offset = 31,
-		.prop_flags = PROP_FLAG_READONLY
-	},
 	{
 		.name = "ToolID",
 		.desc = "Returns the ID of the associated device. ",
@@ -860,6 +590,12 @@ static param_t parameters[] =
 		.prop_offset = 0,
 		.prop_flags = PROP_FLAG_READONLY
 	},
+	{
+		.name = "MapToOutput",
+		.desc = "Map the device to the given output. ",
+		.set_func = set_output,
+		.prop_flags = PROP_FLAG_WRITEONLY
+	},
 	{
 		.name = "all",
 		.desc = "Get value for all parameters.",
@@ -893,6 +629,14 @@ static struct modifier modifiers[] = {
 	{"lshift", "Shift_L"},
 	{"rshift", "Shift_R"},
 
+	{"super", "Super_L"},
+	{"lsuper", "Super_L"},
+	{"rsuper", "Super_R"},
+
+	{"hyper", "Hyper_L"},
+	{"lhyper", "Hyper_L"},
+	{"rhyper", "Hyper_R"},
+
 	{ NULL, NULL }
 };
 
@@ -917,6 +661,8 @@ static struct modifier specialkeys[] = {
 
 	{"backspace", "BackSpace"}, {"Backspace", "BackSpace"},
 
+	{"tab", "Tab"},
+
 	{ NULL, NULL }
 };
 
@@ -964,16 +710,17 @@ static void usage(void)
 	" -h, --help                 - usage\n"
 	" -v, --verbose              - verbose output\n"
 	" -V, --version              - version info\n"
-	" -d, --display disp_name    - override default display\n"
+	" -d, --display \"display\"  - override default display\n"
 	" -s, --shell                - generate shell commands for 'get'\n"
-	" -x, --xconf                - generate X.conf lines for 'get'\n");
+	" -x, --xconf                - generate xorg.conf lines for 'get'\n");
 
 	printf(
 	"\nCommands:\n"
-	" --list [dev|param]           - display known devices, parameters \n"
-	" --list mod                   - display supported modifier and specific keys for keystrokes\n"
-	" --set dev_name param [values...] - set device parameter by name\n"
-	" --get dev_name param [param...] - get current device parameter(s) value by name\n");
+	" --list devices             - display detected devices\n"
+	" --list parameters          - display supported parameters\n"
+	" --list modifiers           - display supported modifier and specific keys for keystrokes\n"
+	" --set \"device name\" parameter [values...] - set device parameter by name\n"
+	" --get \"device name\" parameter [param...]  - get current device parameter(s) value by name\n");
 }
 
 
@@ -1101,7 +848,9 @@ static void list_one_device(Display *dpy, XDeviceInfo *info)
 			if (nitems)
 			{
 				type_name = XGetAtomName(dpy, *(Atom*)data);
-				printf("%-16s %-10s\n", info->name, type_name);
+				printf("%-32s	id: %ld	type: %-10s\n",
+						info->name, info->id,
+						type_name);
 			}
 
 			XFree(data);
@@ -1172,11 +921,14 @@ static void list(Display *dpy, int argc, char **argv)
 	TRACE("'list' requested.\n");
 	if (argc == 0)
 		list_devices(dpy);
-	else if (strcmp(argv[0], "dev") == 0)
+	else if (strcmp(argv[0], "dev") == 0 ||
+		 strcmp(argv[0], "devices") == 0)
 		list_devices(dpy);
-	else if (strcmp(argv[0], "param") == 0)
+	else if (strcmp(argv[0], "param") == 0 ||
+		 strcmp(argv[0], "parameters") == 0)
 		list_param(dpy);
-	else if (strcmp(argv[0], "mod") == 0)
+	else if (strcmp(argv[0], "mod") == 0 ||
+		 strcmp(argv[0], "modifiers") == 0)
 		list_mod(dpy);
 	else
 		printf("unknown argument to list.\n");
@@ -1488,50 +1240,109 @@ static int get_button_number_from_string(const char* string)
 	return atoi(&string[strlen("Button")]);
 }
 
-/* Handles complex button mappings through button actions. */
-static void special_map_buttons(Display *dpy, XDevice *dev, param_t* param, int argc, char **argv)
+static const char *wheel_act_prop[] = {
+	"Wacom Rel Wheel Up Action",
+	"Wacom Rel Wheel Down Action",
+	"Wacom Abs Wheel Up Action",
+	"Wacom Abs Wheel Down Action",
+};
+
+/**
+ * Convert the given property from an 8 bit integer property into an action
+ * atom property. In the default case, this means that a property with
+ * values "4 5 4 5" ends up to have the values
+ * "Wacom RHU Action" "Wacom RHW Action" "Wacom AWU Action" "Wacom AWD
+ * Action"
+ * with each of the properties having :
+ * AC_BUTTON | AC_KEYBTNPRESS | 4 (or 5)
+ * AC_BUTTON | 4 (or 5)
+ *
+ * return 0 on success or 1 on failure.
+ */
+static int convert_wheel_prop(Display *dpy, XDevice *dev, Atom btnact_prop)
 {
-	Atom btnact_prop, prop;
-	unsigned long *data, *btnact_data;
-	int slen = strlen("Button");
-	int btn_no;
+	int i;
 	Atom type;
 	int format;
+	unsigned long btnact_nitems, bytes_after;
+	unsigned char *btnact_data; /* current values (button mappings) */
+	unsigned long *btnact_new_data; /* new values (action atoms) */
+
+	XGetDeviceProperty(dpy, dev, btnact_prop, 0, 100, False,
+				AnyPropertyType, &type, &format, &btnact_nitems,
+				&bytes_after, (unsigned char**)&btnact_data);
+
+	btnact_new_data = calloc(btnact_nitems, sizeof(Atom));
+	if (!btnact_new_data)
+		return 1;
+
+	for (i = 0; i < btnact_nitems; i++) {
+		unsigned long action_data[2];
+		Atom prop = XInternAtom(dpy, wheel_act_prop[i], False);
+
+		action_data[0] = AC_BUTTON | AC_KEYBTNPRESS | btnact_data[i];
+		action_data[1] = AC_BUTTON | btnact_data[i];
+
+		XChangeDeviceProperty(dpy, dev, prop, XA_INTEGER, 32,
+				      PropModeReplace,
+				      (unsigned char*)action_data, 2);
+
+		btnact_new_data[i] = prop;
+	}
+
+	XChangeDeviceProperty(dpy, dev, btnact_prop, XA_ATOM, 32,
+				PropModeReplace,
+				(unsigned char*)btnact_new_data, btnact_nitems);
+	return 0;
+}
+
+
+static void special_map_property(Display *dpy, XDevice *dev, Atom btnact_prop, int offset, int argc, char **argv)
+{
+	unsigned long *data, *btnact_data;
+	Atom type, prop = 0;
+	int format;
 	unsigned long btnact_nitems, nitems, bytes_after;
 	int need_update = 0;
 	int i;
 	int nwords = 0;
 	char **words = NULL;
 
-	TRACE("Special %s map for device %ld.\n", param->name, dev->device_id);
-
-	if (slen >= strlen(param->name) || strncmp(param->name, "Button", slen))
-		return;
-
-	btnact_prop = XInternAtom(dpy, "Wacom Button Actions", True);
-	if (!btnact_prop)
-		return;
-
-	btn_no = get_button_number_from_string(param->name);
-	btn_no--; /* property is zero-indexed, button numbers are 1-indexed */
-
 	XGetDeviceProperty(dpy, dev, btnact_prop, 0, 100, False,
 				AnyPropertyType, &type, &format, &btnact_nitems,
 				&bytes_after, (unsigned char**)&btnact_data);
 
-	if (btn_no > btnact_nitems)
+	if (offset > btnact_nitems)
 		return;
 
-	/* some atom already assigned, modify that */
-	if (btnact_data[btn_no])
-		prop = btnact_data[btn_no];
+	/* Prop is currently 8 bit integer, i.e. plain button
+	 * mappings. Convert to 32 bit Atom actions first.
+	 */
+	if (format == 8 && type == XA_INTEGER)
+	{
+		if (convert_wheel_prop(dpy, dev, btnact_prop))
+			return;
+
+		XGetDeviceProperty(dpy, dev, btnact_prop, 0, 100, False,
+				   AnyPropertyType, &type, &format,
+				   &btnact_nitems, &bytes_after,
+				   (unsigned char**)&btnact_data);
+	}
+
+	if (argc == 0) /* unset property */
+	{
+		prop = btnact_data[offset];
+		btnact_data[offset] = 0;
+	} else if (btnact_data[offset])
+		/* some atom already assigned, modify that */
+		prop = btnact_data[offset];
 	else
 	{
 		char buff[64];
-		sprintf(buff, "Wacom button action %d", (btn_no + 1));
+		sprintf(buff, "Wacom button action %d", (offset + 1));
 		prop = XInternAtom(dpy, buff, False);
 
-		btnact_data[btn_no] = prop;
+		btnact_data[offset] = prop;
 		need_update = 1;
 	}
 
@@ -1560,19 +1371,73 @@ static void special_map_buttons(Display *dpy, XDevice *dev, param_t* param, int
 		}
 	}
 
-	XChangeDeviceProperty(dpy, dev, prop, XA_INTEGER, 32,
-				PropModeReplace,
-				(unsigned char*)data, nitems);
-
-	if (need_update)
-		XChangeDeviceProperty(dpy, dev, btnact_prop, XA_ATOM, 32,
+	if (argc > 0) /* unset property */
+		XChangeDeviceProperty(dpy, dev, prop, XA_INTEGER, 32,
 					PropModeReplace,
-					(unsigned char*)btnact_data,
-					btnact_nitems);
+					(unsigned char*)data, nitems);
+
+	XChangeDeviceProperty(dpy, dev, btnact_prop, XA_ATOM, 32,
+				PropModeReplace,
+				(unsigned char*)btnact_data,
+				btnact_nitems);
+
+	if (argc == 0 && prop)
+		XDeleteDeviceProperty(dpy, dev, prop);
+
 	XFlush(dpy);
 }
 
 
+static void special_map_wheels(Display *dpy, XDevice *dev, param_t* param, int argc, char **argv)
+{
+	Atom wheel_prop;
+
+	wheel_prop = XInternAtom(dpy, param->prop_name, True);
+	if (!wheel_prop)
+		return;
+
+	TRACE("Wheel property %s (%ld)\n", param->prop_name, wheel_prop);
+
+	special_map_property(dpy, dev, wheel_prop, param->prop_offset, argc, argv);
+}
+
+static void map_wheels(Display *dpy, XDevice *dev, param_t* param, int argc, char **argv)
+{
+	if (argc <= 0)
+		return;
+
+	TRACE("Mapping wheel %s for device %ld.\n", param->name, dev->device_id);
+
+	/* FIXME:
+	   if value is simple number, change back to 8 bit integer
+	 */
+
+	special_map_wheels(dpy, dev, param, argc, argv);
+}
+
+/* Handles complex button mappings through button actions. */
+static void special_map_buttons(Display *dpy, XDevice *dev, param_t* param, int argc, char **argv)
+{
+	Atom btnact_prop;
+	int slen = strlen("Button");
+	int btn_no;
+
+	TRACE("Special %s map for device %ld.\n", param->name, dev->device_id);
+
+	if (slen >= strlen(param->name) || strncmp(param->name, "Button", slen))
+		return;
+
+	btnact_prop = XInternAtom(dpy, "Wacom Button Actions", True);
+	if (!btnact_prop)
+		return;
+
+	btn_no = get_button_number_from_string(param->name);
+	btn_no--; /* property is zero-indexed, button numbers are 1-indexed */
+
+	special_map_property(dpy, dev, btnact_prop, btn_no, argc, argv);
+}
+
+
 static void map_button_simple(Display *dpy, XDevice *dev, param_t* param, int button)
 {
 	int nmap = 256;
@@ -1584,18 +1449,23 @@ static void map_button_simple(Display *dpy, XDevice *dev, param_t* param, int bu
 		return;
 
 	nmap = XGetDeviceButtonMapping(dpy, dev, map, nmap);
-	if (btn_no >= nmap)
+	if (btn_no > nmap)
 	{
 		fprintf(stderr, "Button number does not exist on device.\n");
 		return;
 	}
 
+	TRACE("Mapping button %d to %d.\n", btn_no, button);
+
 	map[btn_no - 1] = button;
 	XSetDeviceButtonMapping(dpy, dev, map, nmap);
 	XFlush(dpy);
+
+	/* If there's a property set, unset it */
+	special_map_buttons(dpy, dev, param, 0, NULL);
 }
 /*
-   Supports three variations.
+   Supports two variations, simple mapping and special mapping:
    xsetwacom set device Button1 1
 	- maps button 1 to logical button 1
    xsetwacom set device Button1 "key a b c d"
@@ -1739,58 +1609,6 @@ error:
 	return;
 }
 
-static void set_twinview(Display *dpy, XDevice *dev, param_t* param, int argc, char **argv)
-{
-	int twinview = 0;
-	Atom prop, type;
-	int format;
-	unsigned char* data;
-	unsigned long nitems, bytes_after;
-
-	if (argc != 1)
-		goto error;
-
-	TRACE("TwinView '%s' for device %ld.\n", argv[0], dev->device_id);
-
-	if (strcasecmp(argv[0], "none") == 0)
-		twinview = TV_NONE;
-	else if (strcasecmp(argv[0], "horizontal") == 0)
-		twinview = TV_LEFT_RIGHT;
-	else if (strcasecmp(argv[0], "vertical") == 0)
-		twinview = TV_ABOVE_BELOW;
-	else
-		goto error;
-
-	prop = XInternAtom(dpy, param->prop_name, True);
-	if (!prop)
-	{
-		fprintf(stderr, "Property for '%s' not available.\n",
-			param->name);
-		return;
-	}
-
-	XGetDeviceProperty(dpy, dev, prop, 0, 1000, False, AnyPropertyType,
-				&type, &format, &nitems, &bytes_after, &data);
-
-	if (nitems == 0 || format != 8)
-	{
-		fprintf(stderr, "Property for '%s' has no or wrong value - this is a bug.\n",
-			param->name);
-		return;
-	}
-
-	data[param->prop_offset] = twinview;
-	XChangeDeviceProperty(dpy, dev, prop, type, format,
-				PropModeReplace, data, nitems);
-	XFlush(dpy);
-
-	return;
-
-error:
-	fprintf(stderr, "Usage: xsetwacom rotate <device name> [NONE | CW | CCW | HALF]\n");
-	return;
-}
-
 static int convert_value_from_user(param_t *param, char *value)
 {
 	int val;
@@ -1995,48 +1813,6 @@ static void get_rotate(Display *dpy, XDevice *dev, param_t* param, int argc, cha
 	return;
 }
 
-static void get_twinview(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv)
-{
-	char *twinview = NULL;
-	Atom prop, type;
-	int format;
-	unsigned char* data;
-	unsigned long nitems, bytes_after;
-
-	prop = XInternAtom(dpy, param->prop_name, True);
-	if (!prop)
-	{
-		fprintf(stderr, "Property for '%s' not available.\n",
-			param->name);
-		return;
-	}
-
-	TRACE("Getting twinview setting for device %ld.\n", dev->device_id);
-
-	XGetDeviceProperty(dpy, dev, prop, 0, 1000, False, AnyPropertyType,
-				&type, &format, &nitems, &bytes_after, &data);
-
-	if (nitems == 0 || format != 8)
-	{
-		fprintf(stderr, "Property for '%s' has no or wrong value - this is a bug.\n",
-			param->name);
-		return;
-	}
-
-	switch(data[param->prop_offset])
-	{
-		case TV_NONE: twinview = "none"; break;
-		case TV_ABOVE_BELOW: twinview = "vertical"; break;
-		case TV_LEFT_RIGHT: twinview = "horizontal"; break;
-		default:
-				    break;
-	}
-
-	print_value(param, "%s", twinview);
-
-	return;
-}
-
 static void get_presscurve(Display *dpy, XDevice *dev, param_t *param, int argc,
 				char **argv)
 {
@@ -2072,6 +1848,104 @@ static void get_presscurve(Display *dpy, XDevice *dev, param_t *param, int argc,
 	print_value(param, "%s", buff);
 }
 
+static int get_special_button_map(Display *dpy, XDevice *dev,
+				  param_t *param, int btn_no)
+{
+	Atom btnact_prop, action_prop;
+	unsigned long *btnact_data;
+	Atom type;
+	int format;
+	unsigned long btnact_nitems, bytes_after;
+	int i;
+	char buff[1024] = {0};
+
+	btnact_prop = XInternAtom(dpy, "Wacom Button Actions", True);
+
+	if (!btnact_prop)
+		return 0;
+
+	XGetDeviceProperty(dpy, dev, btnact_prop, 0, 100, False,
+			   AnyPropertyType, &type, &format, &btnact_nitems,
+			   &bytes_after, (unsigned char**)&btnact_data);
+
+	/* button numbers start at 1, property is zero-indexed */
+	if (btn_no >= btnact_nitems)
+		return 0;
+
+	/* FIXME: doesn't cover wheels/strips at the moment, they can be 8
+	 * bits (plain buttons) or 32 bits (complex actions) */
+
+	action_prop = btnact_data[btn_no - 1];
+	if (!action_prop)
+		return 0;
+
+	XFree(btnact_data);
+
+	XGetDeviceProperty(dpy, dev, action_prop, 0, 100, False,
+			   AnyPropertyType, &type, &format, &btnact_nitems,
+			   &bytes_after, (unsigned char**)&btnact_data);
+
+	if (format != 32 && type != XA_ATOM)
+		return 0;
+
+	for (i = 0; i < btnact_nitems; i++)
+	{
+		static int last_type, last_press;
+		unsigned long action = btnact_data[i];
+		int current_type;
+		int detail;
+		int is_press = -1;
+		char str[32] = {0};
+		char press_str = ' ';
+
+		current_type = action & AC_TYPE;
+		detail = action & AC_CODE;
+
+
+		switch (current_type)
+		{
+			case AC_KEY:
+				if (last_type != current_type)
+					strcat(buff, "key ");
+				is_press = !!(action & AC_KEYBTNPRESS);
+				detail = XKeycodeToKeysym(dpy, detail, 0);
+				break;
+			case AC_BUTTON:
+				if (last_type != current_type)
+					strcat(buff, "button ");
+				is_press = !!(action & AC_KEYBTNPRESS);
+				break;
+			case AC_MODETOGGLE:
+				strcat(buff, "modetoggle ");
+				break;
+			case AC_DISPLAYTOGGLE:
+				strcat(buff, "displaytoggle ");
+				break;
+			default:
+				TRACE("unknown type %d\n", current_type);
+				continue;
+		}
+
+		press_str = (is_press == -1) ? ' ' : ((is_press) ?  '+' : '-');
+		if (current_type == AC_KEY)
+			sprintf(str, "%c%s ", press_str,
+				XKeysymToString(detail));
+		else
+			sprintf(str, "%c%d ", press_str, detail);
+		strcat(buff, str);
+		last_type = current_type;
+		last_press = is_press;
+	}
+
+	TRACE("%s\n", buff);
+
+	XFree(btnact_data);
+
+	print_value(param, "%s", buff);
+
+	return 1;
+}
+
 static void get_button(Display *dpy, XDevice *dev, param_t *param, int argc,
 			char **argv)
 {
@@ -2083,7 +1957,11 @@ static void get_button(Display *dpy, XDevice *dev, param_t *param, int argc,
 	if (btn_no == -1)
 		return;
 
-	TRACE("Getting button map curve for device %ld.\n", dev->device_id);
+	TRACE("Getting button map for device %ld.\n", dev->device_id);
+
+	/* if there's a special map, print it and return */
+	if (get_special_button_map(dpy, dev, param, btn_no))
+		return;
 
 	nmap = XGetDeviceButtonMapping(dpy, dev, map, nmap);
 
@@ -2099,13 +1977,133 @@ static void get_button(Display *dpy, XDevice *dev, param_t *param, int argc,
 	XFlush(dpy);
 }
 
+static void _set_matrix_prop(Display *dpy, XDevice *dev, const float fmatrix[9])
+{
+	Atom matrix_prop = XInternAtom(dpy, "Coordinate Transformation Matrix", True);
+	Atom type;
+	int format;
+	unsigned long nitems, bytes_after;
+	float *data;
+	long matrix[9] = {0};
+	int i;
+
+	if (!matrix_prop)
+	{
+		fprintf(stderr, "Server does not support transformation");
+		return;
+	}
+
+	/* XI1 expects 32 bit properties (including float) as long,
+	 * regardless of architecture */
+	for (i = 0; i < sizeof(matrix)/sizeof(matrix[0]); i++)
+		*(float*)(matrix + i) = fmatrix[i];
+
+	XGetDeviceProperty(dpy, dev, matrix_prop, 0, 9, False,
+				AnyPropertyType, &type, &format, &nitems,
+				&bytes_after, (unsigned char**)&data);
+
+	if (format != 32 || type != XInternAtom(dpy, "FLOAT", True))
+		return;
+
+	XChangeDeviceProperty(dpy, dev, matrix_prop, type, format,
+			      PropModeReplace, (unsigned char*)matrix, 9);
+	XFree(data);
+	XFlush(dpy);
+}
+
+static void set_output(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv)
+{
+	int min, maj;
+	int i, found = 0;
+	char *output_name;
+	XRRScreenResources *res;
+	XRROutputInfo *output_info;
+	XRRCrtcInfo *crtc_info;
+
+	output_name = argv[0];
+
+	if (!XRRQueryExtension(dpy, &maj, &min)) /* using min/maj as dummy */
+	{
+		fprintf(stderr, "Server does not support RandR");
+		return;
+	}
+
+	if (!XRRQueryVersion(dpy, &maj, &min) ||
+	    (maj * 1000 + min) < 1002)
+	{
+		fprintf(stderr, "Server does not support RandR 1.2");
+		return;
+	}
+
+
+	res = XRRGetScreenResources(dpy, DefaultRootWindow(dpy));
+
+	for (i = 0; i < res->noutput && !found; i++)
+	{
+		output_info = XRRGetOutputInfo(dpy, res, res->outputs[i]);
+
+		TRACE("Found output '%s' (%s)\n", output_info->name,
+		      output_info->connection == RR_Connected ? "connected" : "disconnnected");
+
+		if (!output_info->crtc || output_info->connection != RR_Connected)
+			continue;
+
+		crtc_info = XRRGetCrtcInfo (dpy, res, output_info->crtc);
+		TRACE("CRTC (%dx%d) %dx%d\n", crtc_info->x, crtc_info->y,
+			crtc_info->width, crtc_info->height);
+
+		if (strcmp(output_info->name, output_name) == 0)
+		{
+			found = 1;
+			break;
+		}
+	}
+
+	/* crtc holds our screen info, need to compare to actual screen size */
+	if (found)
+	{
+		int width = DisplayWidth(dpy, DefaultScreen(dpy));
+		int height = DisplayHeight(dpy, DefaultScreen(dpy));
+
+		/* offset */
+		float x = 1.0 * crtc_info->x/width;
+		float y = 1.0 * crtc_info->y/height;
+
+		/* mapping */
+		float w = 1.0 * crtc_info->width/width;
+		float h = 1.0 * crtc_info->height/height;
+
+		float matrix[9] = { 1, 0, 0,
+				    0, 1, 0,
+				    0, 0, 1};
+		matrix[2] = x;
+		matrix[5] = y;
+		matrix[0] = w;
+		matrix[4] = h;
+
+		TRACE("Transformation matrix:\n");
+		TRACE("	[ %f %f %f ]\n", matrix[0], matrix[1], matrix[2]);
+		TRACE("	[ %f %f %f ]\n", matrix[3], matrix[4], matrix[5]);
+		TRACE("	[ %f %f %f ]\n", matrix[6], matrix[7], matrix[8]);
+
+		_set_matrix_prop(dpy, dev, matrix);
+	} else
+		printf("Unable to find output '%s'. "
+			"Output may not be connected.\n", output_name);
+
+	XRRFreeScreenResources(res);
+}
+
 static void get_all(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv)
 {
 	param_t *p = parameters;
 
+	if (param->printformat == FORMAT_DEFAULT)
+		param->printformat = FORMAT_XORG_CONF;
+
 	while(p->name)
 	{
-		if (p != param)
+		if (p != param && !(p->prop_flags & PROP_FLAG_WRITEONLY))
 		{
 			p->device_name = param->device_name;
 			p->printformat = param->printformat;
@@ -2161,6 +2159,8 @@ static void get_param(Display *dpy, XDevice *dev, param_t *param, int argc, char
 	int format;
 	unsigned char* data;
 	unsigned long nitems, bytes_after;
+	int i;
+	char str[100] = {0};
 
 	if (param->prop_name)
 	{
@@ -2192,23 +2192,31 @@ static void get_param(Display *dpy, XDevice *dev, param_t *param, int argc, char
 	switch(param->prop_format)
 	{
 		case 8:
+			for (i = 0; i < 1 + param->prop_extra; i++)
 			{
-				char str[10] = {0};
-				int val = data[param->prop_offset];
+				int val = data[param->prop_offset + i];
 
 				if (param->prop_flags & PROP_FLAG_BOOLEAN)
-					sprintf(str, "%s", val ?  "on" : "off");
+					sprintf(&str[strlen(str)], "%s", val ?  "on" : "off");
 				else
-					sprintf(str, "%d", val);
-				print_value(param, "%s", str);
+					sprintf(&str[strlen(str)], "%d", val);
+
+				if (i < param->prop_extra)
+					strcat(str, " ");
 			}
+			print_value(param, "%s", str);
 			break;
 		case 32:
+			for (i = 0; i < 1 + param->prop_extra; i++)
 			{
 				long *ldata = (long*)data;
-				print_value(param, "%ld", ldata[param->prop_offset]);
-				break;
+				sprintf(&str[strlen(str)], "%ld", ldata[param->prop_offset + i]);
+
+				if (i < param->prop_extra)
+					strcat(str, " ");
 			}
+			print_value(param, "%s", str);
+			break;
 	}
 }
 
@@ -2329,4 +2337,4 @@ int main (int argc, char **argv)
 }
 
 
-/* vim: set noexpandtab shiftwidth=8: */
+/* vim: set noexpandtab tabstop=8 shiftwidth=8: */
